"""
Tic Tac Toe Player

Resources used:
 - lecture notes: https://cs50.harvard.edu/ai/2020/notes/0/ specifically, the pseudo code for the minimax
 - I also made use of https://stackoverflow.com/questions/27168851/python-sort-a-list-containing-sets?rq=1 to figure out how to sort a list of sets

"""

#  from audioop import reverse
import math
import copy
from unittest.util import sorted_list_difference

X = "X"
O = "O"
EMPTY = None


def initial_state():
    """
    Returns starting state of the board.
    """
    return [[EMPTY, EMPTY, EMPTY],
            [EMPTY, EMPTY, EMPTY],
            [EMPTY, EMPTY, EMPTY]]



def player(board):
    """
    Returns player who has the next turn on a board.
    """
    
    x_count = 0
    o_count = 0

    # loop through board, and count
    for row in range(0, len(board)):
        for column in range(0, len(board[0])):
            if board[row][column] == X:
                x_count += 1
            elif board[row][column] == O:
                o_count += 1
    
    if x_count > o_count:
        return O 
    else:
        return X
    


def actions(board):
    """
    Returns set of all possible actions (i, j) available on the board.
    """
    # create empty set of actions
    actions = set()

    # iterate through list of lists and check if == EMPTY
    for row in range(3):
        for column in range(3):
            if board[row][column] == EMPTY:
                actions.add((row, column))

    # once completed the loop through the board, return actions
    return actions



def result(board, action):
    """
    Returns the board that results from making move (i, j) on the board.
    """
    
    # check if action is valid (= if its among those returned from the actions function)
    if action not in actions(board):
        raise Exception('Invalid action proposed.')  
    
    # make an editable new deep copy of the board first
    new_board = copy.deepcopy(board)

    # add action to new board
    if player(board) == X:
        new_board[action[0]][action[1]] = X
    else:
        new_board[action[0]][action[1]] = O

    # return result
    return new_board
    


def winner(board):
    """
    Returns the winner of the game, if there is one.
    """
    # winner variable
    winner = None

    # check rows:
    for i in range(3):
        j=0
        # find out which player has a chance (which mark is in the first field)
        if board[i][j] != "EMPTY":
            checking = board[i][j]
            # see if that player has marks in entire row
            if board[i][j+1] == checking:
                if board [i][j+2] == checking:
                    winner = checking
    
    # check columns:
    if winner == None:
        for j in range(3):
            i=0
            # find out which player has a chance (which mark is in the first field)
            if board[i][j] != "EMPTY":
                checking = board[i][j]
                # see if that player has marks in entire row
                if board[i+1][j] == checking:
                    if board [i+2][j] == checking:
                        winner = checking

    # check diagonal 1:
    if winner == None:
        j=0
        i=0
        # find out which player has a chance (is in the first field)
        if board[i][j] != "EMPTY":
            checking = board[i][j]
            # see if that player has marks in entire row
            if board[i+1][j+1] == checking:
                if board [i+2][j+2] == checking:
                    winner = checking

    # check diagonal 2:
    if winner == None:
        j=2
        i=0
        # find out which player has a chance (is in the first field)
        if board[i][j] != "EMPTY":
            checking = board[i][j]
            # see if that player has marks in entire row
            if board[i+1][j-1] == checking:
                if board [i+2][j-2] == checking:
                    winner = checking
    
    # check and return result
    if winner == None:
        return None
    if winner == X:
        return X
    if winner == O:
        return O



def terminal(board):
    """
    Returns True if game is over, False otherwise.
    """
    result = None

    # check if there's a winner
    if winner(board) == X: 
        return True
    elif winner(board) == O:
        return True

    # check if there are no EMPTY fields
    for i in range (3):
        for j in range (3):
            if board[i][j] == EMPTY: 
                return False

    # if no empty fields were found, board is full
    if result == None:
        return True
    
    

def utility(board):
    """
    Returns 1 if X has won the game, -1 if O has won, 0 otherwise.
    """
    # this function only gets called if terminal() is True
    if winner(board) == X:
        return 1
    elif winner(board) == O:
        return -1
    else:
        return 0



def minimax(board):
    """
    Returns the optimal action for the current player on the board.
    """
    # input is the board
    # returns optimal move for the player to move: return move = (i, j)
    
    # check if board is terminal (if so: return none)
    if terminal(board) == True:
        return None
    
    
    else:
        # set up new variable "options" to keep track of possible moves and their associated utility
        options = []

        if player(board) == X:
            # maximising player picks action a in Actions(s) that produces the highest value of 
            # Min-Value(Results(s, a))
            for action in actions(board):
                # I am adding the minimum value and the associated action to the list of options
                options.append((min_value(result(board, action)), action))
            
            # in the end, choose the one with the highest min-value. Reverse = true is important here
            sorted_options = sorted(options, key=lambda x: x[0], reverse=True)
            
            # once sorted, the decided move should be in the first set and the second item within that set
            decided_move = sorted_options[0][1]
            return decided_move


        if player(board) == O:
            # minimising player picks action a in Actions(s) that produces the lowest value of 
            # Max-Value(Results(s, a))
            for action in actions(board):
                # I am adding the minimum value and the associated action to the list of options
                options.append((max_value(result(board, action)), action))
            
            # in the end, choose the one with the lowest max-value
            sorted_options = sorted(options, key=lambda x: x[0])
            
            # once sorted, the decided move should be in the first set and the second item within that set
            decided_move = sorted_options[0][1]
            return decided_move



def max_value(board):
    """
    returns the maximum value possible with the given board
    """
    
    # set variable to lower than possible
    v = -math.inf
    
    # breakpoint to the recursion is the terminal state
    if terminal(board) == True:
        return utility(board)
    
    # if no breakpoint, then get all possible actions and check for the one with the highest value
    # store the action that returns the highest value in "v"
    for action in actions(board):
        try_action = max(v, min_value(result(board, action)))
        if try_action > v:
            v = try_action
            # insert a way of alpha beta pruning: if v is at any point identified as "1" then we already have the max value
            if v == 1:
                return v
    return v



def min_value(board):
    """
    returns the minimum value possible with the given board
    """
    
    # set variable to lower than possible
    v = math.inf

    # breakpoint to the recursion is the terminal state
    if terminal(board) == True:
        return utility(board)
    
    # if no breakpoint, then get all possible actions and check for the one with the lowest value
    # store the action that returns the lowest value in "v"
    for action in actions(board):
        try_action = min(v, max_value(result(board, action)))
        if try_action < v:
            v = try_action
            # insert a way of alpha beta pruning: if v is at any point identified as "-1" then we already have the min value
            if v == -1:
                return v

    return v